#ifndef __TWI_H__
#define __TWI_H__

#define TWCR_CMD_MASK 0x0F

void TWI_Init(unsigned long SysClock)
{  
   outb(PORTC, 0x30); /* turn on pullup resistors */
   outb(TWBR, (SysClock / 100000UL - 16) / 2);
   outb(TWSR, 0x01); /* set prescaler */
      
   outb(TWCR, _BV(TWEN) | _BV(TWEA));
   outb(TWAR, 0x01); // only needed for slave mode
}     

char TWI_Stop()
{
   outb(TWCR, (inb(TWCR) & TWCR_CMD_MASK) | BV(TWINT)|BV(TWEA)|BV(TWSTO));
   loop_until_bit_is_clear(TWCR, TWSTO);
   return(0);
}

char TWI_Start()
{
#if DEBUG
   unsigned char Output[30];
#endif

   outb(TWCR, (inb(TWCR) & TWCR_CMD_MASK) | _BV(TWINT) | _BV(TWSTA));
   loop_until_bit_is_set(TWCR, TWINT); //wait for interrupt flag to get set
   if (TW_STATUS != TW_START)
   {
#if DEBUG
      sprintf(Output, "Start failed: 0x%02x\n\r", TW_STATUS);
      UART_TransmitLine(Output);
#endif
      return -1;  //error occured - should have been start
   }
   return(0);
}

char TWI_RepeatedStart()
{
#if DEBUG
      unsigned char Output[30];
#endif

   TWCR = (inb(TWCR) & TWCR_CMD_MASK) | _BV(TWINT)|_BV(TWSTA);  //send repeated start
   loop_until_bit_is_set(TWCR, TWINT);
   if (TW_STATUS != TW_REP_START)
   {
#if DEBUG
      sprintf(Output, "RepStart failed: 0x%02x\n\r", TW_STATUS);
      UART_TransmitLine(Output);
#endif
      TWI_Stop();
      return -1;
   }
   return 0;
}

char TWI_Send(unsigned char SendVal, unsigned char ExpectedACK)
{
#if DEBUG
   unsigned char Output[30];
#endif

   outb(TWDR, SendVal & 0xFE);
   outb(TWCR, (inb(TWCR) & TWCR_CMD_MASK) | _BV(TWINT)); //clear interrupt to start transmission
   loop_until_bit_is_set(TWCR, TWINT); //wait for interrupt flag to get set
   if (TW_STATUS != ExpectedACK)
   {
#if DEBUG
      sprintf(Output, "Error: 0x%02x, ", TW_STATUS);
      UART_TransmitLine(Output);
      sprintf(Output, "Expected: 0x%02x\n\r", ExpectedACK);
      UART_TransmitLine(Output);
#endif
      return -1;
   }
   return(0);
}

char TWI_Receive(unsigned char DevAddr, unsigned char *Data, unsigned char Count)
{
#if DEBUG
   unsigned char Output[30];
#endif
   unsigned char ExpectedACK;
   unsigned char Counter;

   ExpectedACK = TW_MR_SLA_ACK;
   outb(TWDR, DevAddr | 0x01);
   outb(TWCR, (inb(TWCR) & TWCR_CMD_MASK) | _BV(TWINT)); //clear interrupt to start transmission
   loop_until_bit_is_set(TWCR, TWINT); //wait for interrupt flag to get set
   if (TW_STATUS != ExpectedACK)
   {
#if DEBUG
      sprintf(Output, "Error: 0x%02x, ", TW_STATUS);
      UART_TransmitLine(Output);
      sprintf(Output, "Expected: 0x%02x\n\r", ExpectedACK);
      UART_TransmitLine(Output);
#endif
      return -1;
   }

   ExpectedACK = TW_MR_DATA_ACK;
   for (Counter = 0 ; Counter < Count ; Counter ++)
   {
      if (Counter == Count - 1) TWCR = (inb(TWCR) & TWCR_CMD_MASK) | _BV(TWINT);
      else TWCR = (inb(TWCR) & TWCR_CMD_MASK) | _BV(TWINT) | _BV(TWEA);
      loop_until_bit_is_set(TWCR, TWINT);
      if (TW_STATUS != ExpectedACK)
      {
         if (TW_STATUS == TW_MR_DATA_NACK) break; /* we sent the nack */
#if DEBUG
         sprintf(Output, "Error: 0x%02x, ", TW_STATUS);
         UART_TransmitLine(Output);
         sprintf(Output, "Expected: 0x%02x\n\r", ExpectedACK);
         UART_TransmitLine(Output);
#endif
         return -1;
      }
      Data[Counter] = TWDR;
   }
   return(0);
}

char TWI_SendCommand(unsigned char DevAddress, unsigned char Register, unsigned char *Data, unsigned char Count)
{
   char Result;

   loop_until_bit_is_clear(TWCR, TWSTO);  //check no transmission in progress

   Result = TWI_Start();
   if (Result) return -1;

   Result = TWI_Send(DevAddress, TW_MT_SLA_ACK);
   if (Result)
   {
      TWI_Stop();
      return -2;
   }

   Result = TWI_Send(Register, TW_MT_DATA_ACK);
   if (Result)
   {
      TWI_Stop();
      return -3;
   }

/* ignoring count right now.  Just sending one character */
   Result = TWI_Send(Data[0], TW_MT_DATA_ACK);
   if (Result)
   {
      TWI_Stop();
      return -4;
   }

   TWI_Stop();
   return 0;
}

char TWI_GetData(unsigned char DevAddress, unsigned char Register, unsigned char *Data, unsigned char Count)
{
   char Result;

   loop_until_bit_is_clear(TWCR, TWSTO); //wait for last transmission to end

   Result = TWI_Start();
   if (Result)
   {
      return -1;
   }

   Result = TWI_Send(DevAddress, TW_MT_SLA_ACK);
   if (Result)
   {
      TWI_Stop();
      return -2;
   }

   Result = TWI_Send(Register, TW_MT_DATA_ACK);
   if (Result)
   {
      TWI_Stop();
      return -3;
   }

   Result = TWI_RepeatedStart();
   if (Result)
   {
      TWI_Stop();
      return -4;
   }

   Result = TWI_Receive(DevAddress, Data, Count);
   if (Result)
   {
      TWI_Stop();
      return -5;
   }

   TWI_Stop();
   return 0;
}

#endif
