#ifndef __SERIAL_H__
#define __SERIAL_H__

#define UART_BUFFER_LENGTH 64

/* UART Buffer Defines */
#define UART_RX_BUFFER_SIZE 32 /* 1,2,4,8,16,32,64,128 or 256 bytes */
#define UART_RX_BUFFER_MASK ( UART_RX_BUFFER_SIZE - 1 )
#if ( UART_RX_BUFFER_SIZE & UART_RX_BUFFER_MASK )
#error RX buffer size is not a power of 2
#endif
#define UART_TX_BUFFER_SIZE 32 /* 1,2,4,8,16,32,64,128 or 256 bytes */
#define UART_TX_BUFFER_MASK ( UART_TX_BUFFER_SIZE - 1 )

/* Static Variables */
static unsigned char UART_RxBuf[UART_RX_BUFFER_SIZE];
static volatile unsigned char UART_RxHead;
static volatile unsigned char UART_RxTail;
static unsigned char UART_TxBuf[UART_TX_BUFFER_SIZE];
static volatile unsigned char UART_TxHead;
static volatile unsigned char UART_TxTail;

/* initialize UART */
void UART_Init( unsigned char baudrate )
{
   unsigned char x;
   outp((unsigned char)(baudrate >> 8), UBRRH);
   outp((unsigned char)(baudrate), UBRRL);
   /* enable UART receiver and transmitter, and receive interrupt */
   outp((1<<RXCIE) | (1<<RXEN) | (1<<TXEN), UCSRB);
   /*   outp((1<<URSEL) | (1<<UCSZ1) | (1<<UCSZ0), UCSRC); */
   outp(0x86, UCSRC);
   x = 0; /* flush receive buffer */
   UART_RxTail = x;
   UART_RxHead = x;
   UART_TxTail = x;
   UART_TxHead = x;
}

/* interrupt handlers */
//interrupt [UART_RX_vect] void UART_RX_interrupt( void )
SIGNAL(SIG_UART_RECV)
{
   unsigned char data;
   unsigned char tmphead;
   data = inp(UDR);
   /* calculate buffer index */
   tmphead = ( UART_RxHead + 1 ) & UART_RX_BUFFER_MASK;
   UART_RxHead = tmphead; /* store new index */
   if ( tmphead == UART_RxTail )
   {
      /* ERROR! Receive buffer overflow */
   }
   UART_RxBuf[tmphead] = data; /* store received data in buffer */
}

//interrupt [UART_UDRE_vect] void UART_TX_interrupt(void)
SIGNAL(SIG_UART_DATA)
{
   unsigned char tmptail;

   /* check if all data is transmitted */
   if ( UART_TxHead != UART_TxTail )
   {
      /* calculate buffer index */
      tmptail = ( UART_TxTail + 1 ) & UART_TX_BUFFER_MASK;
      UART_TxTail = tmptail; /* store new index */
      outp(UART_TxBuf[tmptail], UDR);
//      outp(0xAA, UDR);
   }
   else
   {
//       outp(inp(UCR) & ~(1 << UDRIE), UCR);
      cbi(UCSRB, UDRIE);
   }
}

/* Read and write functions */
void UART_TransmitByte( unsigned char data )
{
   unsigned char tmphead;
   /* calculate buffer index */
   tmphead = ( UART_TxHead + 1 ) & UART_TX_BUFFER_MASK; /* wait for free space in buffer */
   while ( tmphead == UART_TxTail );

   UART_TxBuf[tmphead] = data; /* store data in buffer */
   UART_TxHead = tmphead; /* store new index */
//   outp(inp(UCR) | (1 << UDRIE), UCR);
   sbi(UCSRB, UDRIE);
}

void UART_TransmitLine(unsigned char *TransmitBuffer)
{
   unsigned char BufferCounter = 0;
   while (TransmitBuffer[BufferCounter] != '\0')
   {
      UART_TransmitByte(TransmitBuffer[BufferCounter++]);
   }
}

unsigned char UART_ReceiveByte( void )
{
   unsigned char tmptail;
   while ( UART_RxHead == UART_RxTail ) /* wait for incomming data */
      ;
   tmptail = ( UART_RxTail + 1 ) & UART_RX_BUFFER_MASK;/* calculate buffer index */
   UART_RxTail = tmptail; /* store new index */
   return UART_RxBuf[tmptail]; /* return data */
}

void UART_ReceiveLine(unsigned char *ReturnBuffer) // ReturnBuffer must be at least UART_BUFFER_LENGTH long
{
   unsigned char BufferCounter = 0;
//   unsigned char OutputLine[20];
//   unsigned char TempLine[20];
   while (BufferCounter < UART_BUFFER_LENGTH - 2)
   {
      ReturnBuffer[BufferCounter] = UART_ReceiveByte();
      if ((ReturnBuffer[BufferCounter] == '\n') || (ReturnBuffer[BufferCounter] == '\r'))
      {
         ReturnBuffer[BufferCounter] = '\0';
         break;
      }
      if (((ReturnBuffer[BufferCounter] == '\b') || (ReturnBuffer[BufferCounter] == 127)) && (BufferCounter > 0))
      {
         BufferCounter --;
         UART_TransmitByte('\b');
         continue;
      }
      UART_TransmitByte(ReturnBuffer[BufferCounter]);
      BufferCounter ++;
   }
   ReturnBuffer[BufferCounter++] = '\0';
}

unsigned char UART_Data(void)
{
   return(UART_RxHead != UART_RxTail); /* return 0 (FALSE) if the receive buffer is empty */
}

void UART_clr(void)
{ // Send a 'clear screen' to a VT100 terminal
   UART_TransmitByte(0x1B);
   UART_TransmitByte('[');
   UART_TransmitByte('H');
   UART_TransmitByte(0x1B);
   UART_TransmitByte('[');
   UART_TransmitByte('2');
   UART_TransmitByte('J');
   UART_TransmitByte(0x00);
}


#endif
