#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/signal.h>
#include <stdlib.h>
#include <inttypes.h>
#include <avr/eeprom.h>
#include <string.h>

/* unsigned int PodType; */
#define PodType POD_TYPE_THERMOMETER
#define POD_VAR_URL "http://pel.cs.byu.edu/~sorenson/Thesis/PODs/Thermometer/"

#include "PODDefines.h"
#include "PODIncludes.h"
#include "Serial.h"
#include "Sleep.h"
#include "ADC.h"

//char CompileDate __attribute__ ((section (".eeprom"))) = __DATE__;

/* Local POD Variables (eeprom addresses, lengths, etc.) */
#define POD_VAR_WARN_TYPE_ADDR 0x0040
#define POD_VAR_WARN_TYPE_LEN  0x0001

#define POD_VAR_WARN_TEMP_ADDR 0x0041
#define POD_VAR_WARN_TEMP_LEN  0x0004

#define POD_VAR_UNITS_ADDR     0x0045
#define POD_VAR_UNITS_LEN      0x0001

/* WarningType - Warning Type
 * -1 - Low temperature warning
 * 0  - No warning set (default)
 * 1  - High temperature warning
 * */
//char WarningType __attribute__ ((section (".eeprom")));
char WarningType;

/* WarnTemperature - Temperature of warning (in Fahrenheit hundredths - 50F will show as 5000) */
long WarnTemperature;

/* DisplayUnits - Units to display temperature in
 * 0 - Fahrenheit
 * 1 - Celsius
 * */
unsigned char DisplayUnits;

/* WarningState - whether the system is already in a warning state - internally used */
char WarningState;

/* read variables from the eeprom */
void ReadVars()
{
   WarningType = eeprom_read_byte((uint8_t *)(POD_VAR_WARN_TYPE_ADDR));
   WarnTemperature = eeprom_read_word((uint16_t *)(POD_VAR_WARN_TEMP_ADDR))
      + eeprom_read_word((uint16_t *)(POD_VAR_WARN_TEMP_ADDR + 2));
   DisplayUnits = eeprom_read_byte((uint8_t *)(POD_VAR_UNITS_ADDR));
   if (DisplayUnits != 'F') DisplayUnits = 'C';
   WarningState = 0;
}

int Init()
{
   int Prescaler;

   cli();
   Prescaler = 5;
   UART_Init(23);
   SetupADC(0, 1);

   ReadVars();
/*
   WarningState = 0;
   WarnTemperature = 70;
   WarningType = 0;
   DisplayUnits = 0;
*/

   sei();
   return(0);
}

/* returns the temperature in 100ths F */
unsigned long GetReading()
{
   unsigned long ADCValue;
   unsigned long ADCAccumulation;
   int Counter;

   ADCAccumulation = 0;
   for (Counter = 0 ; Counter < 16 ; Counter ++)
   {
      ADCValue = StartADC(0);
      ADCAccumulation += ADCValue;
   }
   ADCValue = ADCAccumulation / 16;

   ADCValue = (ADCValue * 25600) / 1024;
   return(ADCValue);
}

int SendAlert(int AlertType)
{
#if 0
   UART_TransmitByte(POD_RESPONSE_ALERT);
   if (AlertType == 1) UART_TransmitWord(POD_ALERT_THERMOMETER_TEMP_HIGH);
   else if (AlertType == -1) UART_TransmitWord(POD_ALERT_THERMOMETER_TEMP_LOW);
   else UART_TransmitWord(POD_ALERT_THERMOMETER_TEMP_OKAY);
#endif
   return(0);
}

/* Assumes temperature is in 100ths C */
unsigned long CtoF(unsigned long InputTemp)
{
   unsigned long ReturnValue;
   ReturnValue = ((InputTemp * 9) / 5) + 3200;
   return(ReturnValue);
}

/* Assumes temperature is in 100ths F */
unsigned long FtoC(unsigned long InputTemp)
{
   unsigned long ReturnValue;
   ReturnValue = ((InputTemp - 3200) * 5) / 9;
   return(ReturnValue);
}

/* Assumes temperature coming in as 100ths F */
unsigned long TemperatureConvert(unsigned long InputTemp, char Unit)
{
   if (Unit == 0) return(InputTemp);
   if (Unit == 1) return(FtoC(InputTemp));
   return(0);
}

void Process_Command_GetTemp()
{
   unsigned long TemperatureReading; /* Temperature reading from GetReading (in F) */
   unsigned long CurrentTemperature; /* CurrentTemperature in appropriate units */

   TemperatureReading = GetReading();
   CurrentTemperature = TemperatureConvert(TemperatureReading, DisplayUnits);
   UART_TransmitByte(POD_RESPONSE_ACK);
   UART_TransmitWord(CurrentTemperature / 100);
   UART_TransmitByte(CurrentTemperature % 100); /* send hundredths */
}

void Process_Command_GetUnits()
{
   UART_TransmitByte(POD_RESPONSE_ACK);
   if (DisplayUnits == 0) UART_TransmitByte('F');
   else if (DisplayUnits == 1) UART_TransmitByte('C');
   else UART_TransmitByte('*');
}
void Process_Command_SetUnits()
{
   unsigned char Units;
   Units = UART_ReceiveByte();
   if (Units == 'F') DisplayUnits = 0;
   else DisplayUnits = 1;
   eeprom_write_byte((uint8_t *)(POD_VAR_UNITS_ADDR), DisplayUnits);
   UART_TransmitByte(POD_RESPONSE_ACK);
}
void Process_Command_SetWarn()
{
   unsigned long WarnHigh;
   unsigned char WarnLow;
   unsigned long SetTemp;
   WarnHigh = UART_ReceiveWord();
   WarnLow = UART_ReceiveByte();
   if (DisplayUnits != 0)
   { /* convert temperature to F, then store */
      SetTemp = CtoF((WarnHigh * 100) + WarnLow);
   }
   else SetTemp = (WarnHigh * 100) + WarnLow;
   WarnTemperature = SetTemp;
   WarningState = 0;

   eeprom_write_byte((uint8_t *)(POD_VAR_WARN_TYPE_ADDR), WarningType);
   eeprom_write_word((uint16_t *)(POD_VAR_WARN_TEMP_ADDR), (WarnTemperature >> 16));
   eeprom_write_word((uint16_t *)(POD_VAR_WARN_TEMP_ADDR + 2), (WarnTemperature & 0xFFFF));
   UART_TransmitByte(POD_RESPONSE_ACK);
}
void Process_Command_GetWarn()
{
   unsigned long WarnHigh;
   unsigned char WarnLow;
   unsigned long SetTemp;

   if (DisplayUnits != 0) SetTemp = CtoF(WarnTemperature);
   else SetTemp = WarnTemperature;
   WarnHigh = SetTemp / 100;
   WarnLow = SetTemp % 100;
   UART_TransmitByte(POD_RESPONSE_ACK);
   UART_TransmitWord(WarnHigh);
   UART_TransmitByte(WarnLow);
}

int Process_Command(unsigned int Command)
{
   switch (Command)
   {
      case POD_COMMAND_PING:
         Process_Command_Ping();
         break;
      case POD_COMMAND_GETTYPE:
         Process_Command_GetType();
         break;
      case POD_COMMAND_GETINFO:
         Process_Command_GetInfo();
         break;
      case POD_COMMAND_RESET:
/*         Process_Command_Reset(); */
         break;
      case POD_COMMAND_GETNAME:
         Process_Command_GetName();
         break;
      case POD_COMMAND_SETNAME:
         Process_Command_SetName();
         break;
      case POD_COMMAND_THERMOMETER_GET_TEMP:
         Process_Command_GetTemp();
         break;
      case POD_COMMAND_THERMOMETER_GET_UNITS:
         Process_Command_GetUnits();
         break;
      case POD_COMMAND_THERMOMETER_SET_UNITS:
         Process_Command_SetUnits();
         break;
      case POD_COMMAND_THERMOMETER_SET_WARN:
         Process_Command_SetWarn();
         break;
      case POD_COMMAND_THERMOMETER_GET_WARN:
         Process_Command_GetWarn();
         break;
      default:
         Send_NACK();
         break;
   }
   return(0);
}

int main(int argc, char *argv[])
{
   unsigned long ADCValue;
   long CurrentTemp;
   long LastTemp;
   unsigned int Command;
/*   unsigned char TempBuffer[100]; */

   Init();

   ADCValue = StartADC(0);

//   UART_clr();
//   UART_TransmitLine("Starting1\n\r");
   ADCValue = GetReading();
   LastTemp = GetReading();
   while (1)
   {
      CurrentTemp = GetReading();
      if (CurrentTemp != LastTemp)
      {
         if (WarningType == 1) /* check for over-temperature */
         {
            if (CurrentTemp >= WarnTemperature)
            {
               if (!WarningState)
               {
                  SendAlert(1);
                  WarningState = 1;
               }
            } /* if over temp */
         } /* check for over temp */
         else if (WarningType == -1) /* check for low-temperature */
         {
            if (CurrentTemp <= WarnTemperature)
            {
               if (!WarningState)
               {
                  SendAlert(-1);
                  WarningState = 1;
               }
            } /* if low temp */
         } /* check for low temp */
      } /* temperature has changed */
      if (UART_Data()) /* some data is waiting */
      {
         Command = UART_ReceiveWord();
         Process_Command(Command);
      } /* if UART_Data */
   } /* end while */
   return(0);
}
